@@ -5,35 +5,38 @@ import functools
 import inspect
 import json
 import os
+import posixpath
 import re
 import secrets
 import urllib.parse as urlparse
-import warnings
 from pathlib import Path
 from typing import TYPE_CHECKING, Any, Sequence, TypeVar, cast
 
-from packaging.markers import InvalidMarker
 from packaging.requirements import InvalidRequirement
 from packaging.requirements import Requirement as PackageRequirement
-from packaging.specifiers import SpecifierSet
+from packaging.specifiers import InvalidSpecifier, SpecifierSet
 from packaging.utils import parse_sdist_filename, parse_wheel_filename
-from packaging.version import parse as parse_version
-from unearth import Link
 
 from pdm.compat import Distribution
-from pdm.exceptions import ExtrasWarning, RequirementError
-from pdm.models.markers import Marker, get_marker, split_marker_extras
+from pdm.exceptions import RequirementError
+from pdm.models.backends import BuildBackend, get_relative_path
+from pdm.models.markers import Marker, get_marker
 from pdm.models.setup import Setup
-from pdm.models.specifiers import PySpecSet, get_specifier
+from pdm.models.specifiers import PySpecSet, fix_legacy_specifier, get_specifier
 from pdm.utils import (
+    PACKAGING_22,
     add_ssh_scheme_to_git_uri,
+    comparable_version,
     normalize_name,
     path_to_url,
+    path_without_fragments,
     url_to_path,
     url_without_fragments,
 )
 
 if TYPE_CHECKING:
+    from unearth import Link
+
     from pdm._types import RequirementDict
 
 
@@ -51,6 +54,7 @@ _file_req_re = re.compile(
 )
 _egg_info_re = re.compile(r"([a-z0-9_.]+)-([a-z0-9_.!+-]+)", re.IGNORECASE)
 T = TypeVar("T", bound="Requirement")
+ALLOW_ANY = SpecifierSet()
 
 
 def strip_extras(line: str) -> tuple[str, tuple[str, ...] | None]:
@@ -76,14 +80,13 @@ class Requirement:
     name: str | None = None
     marker: Marker | None = None
     extras: Sequence[str] | None = None
-    specifier: SpecifierSet | None = None
+    specifier: SpecifierSet = ALLOW_ANY
     editable: bool = False
-    prerelease: bool = False
+    prerelease: bool | None = None
+    groups: list[str] = dataclasses.field(default_factory=list)
 
     def __post_init__(self) -> None:
-        self.requires_python = (
-            self.marker.split_pyspec()[1] if self.marker else PySpecSet()
-        )
+        self.requires_python = self.marker.split_pyspec()[1] if self.marker else PySpecSet()
 
     @property
     def project_name(self) -> str | None:
@@ -95,9 +98,6 @@ class Requirement:
 
     @property
     def is_pinned(self) -> bool:
-        if not self.specifier:
-            return False
-
         if len(self.specifier) != 1:
             return False
 
@@ -108,11 +108,7 @@ class Requirement:
         """Return a new requirement with the given pinned version."""
         if self.is_pinned or not other_version:
             return self
-        version = parse_version(other_version)
-        normalized = str(version)
-        if version.local:
-            # Remove the local part to accept wider range of prereleases.
-            normalized = normalized.rsplit("+", 1)[0]
+        normalized = comparable_version(other_version)
         return dataclasses.replace(self, specifier=get_specifier(f"=={normalized}"))
 
     def _hash_key(self) -> tuple:
@@ -143,24 +139,15 @@ class Requirement:
     @classmethod
     def create(cls: type[T], **kwargs: Any) -> T:
         if "marker" in kwargs:
-            try:
-                kwargs["marker"] = get_marker(kwargs["marker"])
-            except InvalidMarker as e:
-                raise RequirementError("Invalid marker: %s" % str(e)) from None
+            kwargs["marker"] = get_marker(kwargs["marker"])
         if "extras" in kwargs and isinstance(kwargs["extras"], str):
-            kwargs["extras"] = tuple(
-                e.strip() for e in kwargs["extras"][1:-1].split(",")
-            )
-        version = kwargs.pop("version", None)
-        if version:
+            kwargs["extras"] = tuple(e.strip() for e in kwargs["extras"][1:-1].split(","))
+        version = kwargs.pop("version", "")
+        try:
             kwargs["specifier"] = get_specifier(version)
-        return cls(
-            **{
-                k: v
-                for k, v in kwargs.items()
-                if k in inspect.signature(cls).parameters
-            }
-        )
+        except InvalidSpecifier as e:
+            raise RequirementError(f'Invalid specifier for {kwargs.get("name")}: {version}: {e}') from None
+        return cls(**{k: v for k, v in kwargs.items() if k in inspect.signature(cls).parameters})
 
     @classmethod
     def from_dist(cls, dist: Distribution) -> Requirement:
@@ -168,7 +155,7 @@ class Requirement:
         if direct_url_json is not None:
             direct_url = json.loads(direct_url_json)
             data = {
-                "name": dist.metadata["Name"],
+                "name": dist.metadata.get("Name"),
                 "url": direct_url.get("url"),
                 "editable": direct_url.get("dir_info", {}).get("editable"),
                 "subdirectory": direct_url.get("subdirectory"),
@@ -182,21 +169,19 @@ class Requirement:
                 )
                 return VcsRequirement.create(**data)
             return FileRequirement.create(**data)
-        return NamedRequirement.create(
-            name=dist.metadata["Name"], version=f"=={dist.version}"
-        )
+        return NamedRequirement.create(name=dist.metadata["Name"], version=f"=={dist.version}")
 
     @classmethod
-    def from_req_dict(cls, name: str, req_dict: RequirementDict) -> Requirement:
+    def from_req_dict(cls, name: str, req_dict: RequirementDict, check_installable: bool = True) -> Requirement:
         if isinstance(req_dict, str):  # Version specifier only.
-            return NamedRequirement(name=name, specifier=get_specifier(req_dict))
+            return NamedRequirement.create(name=name, version=req_dict)
         for vcs in VCS_SCHEMA:
             if vcs in req_dict:
                 repo = cast(str, req_dict.pop(vcs, None))
                 url = f"{vcs}+{repo}"
                 return VcsRequirement.create(name=name, vcs=vcs, url=url, **req_dict)
         if "path" in req_dict or "url" in req_dict:
-            return FileRequirement.create(name=name, **req_dict)
+            return FileRequirement.create(name=name, **req_dict, check_installable=check_installable)
         return NamedRequirement.create(name=name, **req_dict)
 
     @property
@@ -218,14 +203,18 @@ class Requirement:
         """Return whether the passed in PEP 508 string
         is the same requirement as this one.
         """
-        if line.strip().startswith("-e "):
-            req = parse_requirement(line.split("-e ", 1)[-1], True)
-        else:
-            req = parse_requirement(line, False)
-        return self.key == req.key
+        req = parse_line(line)
+        return (
+            self.key == req.key
+            or isinstance(self, FileRequirement)
+            and isinstance(req, FileRequirement)
+            and self.url == req.url
+        )
 
     @classmethod
     def from_pkg_requirement(cls, req: PackageRequirement) -> Requirement:
+        from unearth import Link
+
         kwargs = {
             "name": req.name,
             "extras": req.extras,
@@ -235,13 +224,13 @@ class Requirement:
         if getattr(req, "url", None):
             link = Link(cast(str, req.url))
             klass = VcsRequirement if link.is_vcs else FileRequirement
-            return klass(url=req.url, **kwargs)  # type: ignore
+            return klass(url=cast(str, req.url), **kwargs)  # type: ignore[arg-type]
         else:
-            return NamedRequirement(**kwargs)  # type: ignore
+            return NamedRequirement(**kwargs)  # type: ignore[arg-type]
 
     def _format_marker(self) -> str:
         if self.marker:
-            return f"; {str(self.marker)}"
+            return f"; {self.marker!s}"
         return ""
 
 
@@ -249,7 +238,7 @@ class Requirement:
 class NamedRequirement(Requirement):
     def as_line(self) -> str:
         extras = f"[{','.join(sorted(self.extras))}]" if self.extras else ""
-        return f"{self.project_name}{extras}{self.specifier}{self._format_marker()}"
+        return f"{self.project_name}{extras}{self.specifier or ''}{self._format_marker()}"
 
 
 @dataclasses.dataclass(eq=False)
@@ -257,77 +246,113 @@ class FileRequirement(Requirement):
     url: str = ""
     path: Path | None = None
     subdirectory: str | None = None
+    check_installable: bool = True
+    _root: Path = dataclasses.field(default_factory=Path.cwd, repr=False)
 
     def __post_init__(self) -> None:
         super().__post_init__()
         self._parse_url()
-        if self.path and not self.path.exists():
-            raise RequirementError(f"The local path '{self.path}' does not exist.")
-        if self.is_local_dir:
+        if self.is_local_dir and self.check_installable:
             self._check_installable()
 
     def _hash_key(self) -> tuple:
-        return super()._hash_key() + (self.url, self.editable)
+        return (*super()._hash_key(), self.get_full_url(), self.editable)
+
+    def guess_name(self) -> str | None:
+        filename = os.path.basename(urlparse.unquote(url_without_fragments(self.url))).rsplit("@", 1)[0]
+        if self.is_vcs:
+            if self.vcs == "git":  # type: ignore[attr-defined]
+                name = filename
+                if name.endswith(".git"):
+                    name = name[:-4]
+                return name
+            elif self.vcs == "hg":  # type: ignore[attr-defined]
+                return filename
+            else:  # svn and bzr
+                name, in_branch, _ = filename.rpartition("/branches/")
+                if not in_branch and name.endswith("/trunk"):
+                    return name[:-6]
+                return name
+        elif filename.endswith(".whl"):
+            return parse_wheel_filename(filename)[0]
+        else:
+            try:
+                return parse_sdist_filename(filename)[0]
+            except ValueError:
+                match = _egg_info_re.match(filename)
+                # Filename is like `<name>-<version>.tar.gz`, where name will be
+                # extracted and version will be left to be determined from
+                # the metadata.
+                if match:
+                    return match.group(1)
+        return None
 
     @classmethod
     def create(cls: type[T], **kwargs: Any) -> T:
         if kwargs.get("path"):
-            kwargs["path"] = Path(kwargs["path"].replace("${PROJECT_ROOT}", "."))
-        return super().create(**kwargs)  # type: ignore
+            kwargs["path"] = Path(kwargs["path"])
+        return super().create(**kwargs)
 
     @property
     def str_path(self) -> str | None:
         if not self.path:
             return None
-        result = self.path.as_posix()
-        if not self.path.is_absolute() and not result.startswith(("./", "../")):
+        if self.path.is_absolute():
+            try:
+                result = self.path.relative_to(self._root).as_posix()
+            except ValueError:
+                return self.path.as_posix()
+        else:
+            result = self.path.as_posix()
+        result = posixpath.normpath(result)
+        if not result.startswith(("./", "../")):
             result = "./" + result
         if result.startswith("./../"):
             result = result[2:]
         return result
 
     def _parse_url(self) -> None:
-        if not self.url:
-            if self.path:
-                if not self.path.is_absolute():
-                    str_path = cast(str, self.str_path)
-                    if str_path.startswith("./"):
-                        str_path = str_path[2:]
-                    self.url = f"file:///${{PROJECT_ROOT}}/{str_path}"
-                else:
-                    self.url = path_to_url(self.path.absolute().as_posix())
-        else:
-            try:
-                self.path = Path(
-                    url_to_path(
-                        self.url.replace(
-                            "${PROJECT_ROOT}",
-                            Path(".").absolute().as_posix().lstrip("/"),
-                        )
-                    )
-                )
-            except AssertionError:
-                pass
-        self._parse_name_from_url()
+        if not self.url and self.path and self.path.is_absolute():
+            self.url = path_to_url(self.path.as_posix())
+        if not self.path:
+            path = get_relative_path(self.url)
+            if path is None:
+                try:
+                    self.path = path_without_fragments(url_to_path(self.url))
+                except AssertionError:
+                    pass
+            else:
+                self.path = path_without_fragments(path)
+        if self.url:
+            self._parse_name_from_url()
 
-    def relocate(self, root: str | Path) -> None:
+    def relocate(self, backend: BuildBackend) -> None:
         """Change the project root to the given path"""
         if self.path is None or self.path.is_absolute():
             return
         # self.path is relative
-        self.path = Path(os.path.relpath(self.path, root))
-        self.url = ""
-        self._parse_url()
+        self.path = path_without_fragments(os.path.relpath(self.path, backend.root))
+        path = self.path.as_posix()
+        if path == ".":
+            path = ""
+        self.url = backend.relative_path_to_url(path)
+        self._root = backend.root
+
+    @property
+    def absolute_path(self) -> Path | None:
+        return self._root.joinpath(self.path) if self.path else None
 
     @property
     def is_local(self) -> bool:
-        return self.path and self.path.exists() or False
+        return (path := self.absolute_path) is not None and path.exists()
 
     @property
     def is_local_dir(self) -> bool:
-        return self.is_local and cast(Path, self.path).is_dir()
+        return self.is_local and cast(Path, self.absolute_path).is_dir()
 
     def as_file_link(self) -> Link:
+        from unearth import Link
+
         url = self.get_full_url()
         # only subdirectory is useful in a file link
         if self.subdirectory:
@@ -339,11 +364,7 @@ class FileRequirement(Requirement):
 
     def as_line(self) -> str:
         project_name = f"{self.project_name}" if self.project_name else ""
-        extras = (
-            f"[{','.join(sorted(self.extras))}]"
-            if self.extras and self.project_name
-            else ""
-        )
+        extras = f"[{','.join(sorted(self.extras))}]" if self.extras and self.project_name else ""
         marker = self._format_marker()
         if marker:
             marker = f" {marker}"
@@ -370,28 +391,13 @@ class FileRequirement(Requirement):
             if not self.extras:
                 self.extras = extras
         if not self.name and not self.is_vcs:
-            filename = os.path.basename(
-                urlparse.unquote(url_without_fragments(self.url))
-            )
-            if filename.endswith(".whl"):
-                self.name, *_ = parse_wheel_filename(filename)
-            else:
-                try:
-                    self.name, *_ = parse_sdist_filename(filename)
-                except ValueError:
-                    match = _egg_info_re.match(filename)
-                    # Filename is like `<name>-<version>.tar.gz`, where name will be
-                    # extracted and version will be left to be determined from
-                    # the metadata.
-                    if match:
-                        self.name = match.group(1)
+            self.name = self.guess_name()
 
     def _check_installable(self) -> None:
         assert self.path
-        if not (
-            self.path.joinpath("setup.py").exists()
-            or self.path.joinpath("pyproject.toml").exists()
-        ):
+        if not self.path.exists():
+            return
+        if not (self.path.joinpath("setup.py").exists() or self.path.joinpath("pyproject.toml").exists()):
             raise RequirementError(f"The local path '{self.path}' is not installable.")
         result = Setup.from_directory(self.path.absolute())
         if result.name:
@@ -433,69 +439,87 @@ class VcsRequirement(FileRequirement):
             path, ref = parsed.path.split("@", 1)
         repo = urlparse.urlunparse(parsed._replace(path=path, fragment=""))
         self.url = f"{vcs}+{repo}"
-        self.repo, self.ref = repo, ref  # type: ignore
+        self.repo, self.ref = repo, ref
 
 
 def filter_requirements_with_extras(
-    project_name: str,
-    requirement_lines: list[str],
-    extras: Sequence[str],
-    include_default: bool = False,
-) -> list[str]:
+    requirement_lines: list[str], extras: Sequence[str], include_default: bool = False
+) -> list[Requirement]:
     """Filter the requirements with extras.
     If extras are given, return those with matching extra markers.
     Otherwise, return those without extra markers.
     """
-    result: list[str] = []
-    extras_in_meta: set[str] = set()
+    result: list[Requirement] = []
     for req in requirement_lines:
         _r = parse_requirement(req)
+        req_extras = get_marker("")
         if _r.marker:
-            req_extras, rest = split_marker_extras(str(_r.marker))
-            if req_extras:
-                extras_in_meta.update(req_extras)
-                _r.marker = Marker(rest) if rest else None
-        else:
-            req_extras = set()
-        if (
-            req_extras
-            and not req_extras.isdisjoint(extras)
-            or not req_extras
-            and (include_default or not extras)
-        ):
-            result.append(_r.as_line())
-
-    extras_not_found = [e for e in extras if e not in extras_in_meta]
-    if extras_not_found:
-        warnings.warn(ExtrasWarning(project_name, extras_not_found))
+            rest, req_extras = _r.marker.split_extras()
+            _r.marker = rest if not rest.is_any() else None
+            if not req_extras.evaluate({"extra": extras or ""}):
+                continue
+        # Add to the requirements if:
+        # The requirement has no extras while requested extras are empty or include_default is True, or
+        # The requirement has extras, in which case the `evaluate()` test must have been passed.
+        if not req_extras.is_any() or include_default or not extras:
+            result.append(_r)
 
     return result
 
 
-def parse_requirement(line: str, editable: bool = False) -> Requirement:
+def parse_as_pkg_requirement(line: str) -> PackageRequirement:
+    """Parse a requirement line as packaging.requirement.Requirement"""
+    try:
+        return PackageRequirement(line)
+    except InvalidRequirement:
+        if not PACKAGING_22:  # We can't do anything, reraise the error.
+            raise
+        new_line = fix_legacy_specifier(line)
+        return PackageRequirement(new_line)
 
+
+def parse_line(line: str) -> Requirement:
+    if line.startswith("-e "):
+        return parse_requirement(line[3:].strip(), editable=True)
+    return parse_requirement(line)
+
+
+def parse_requirement(line: str, editable: bool = False) -> Requirement:
     m = _vcs_req_re.match(line)
     r: Requirement
     if m is not None:
         r = VcsRequirement.create(**m.groupdict())
     else:
+        # Special handling for hatch local references:
+        # https://hatch.pypa.io/latest/config/dependency/#local
+        # We replace the {root.uri} temporarily with a dummy URL header
+        # to make it pass through the packaging.requirement parser
+        # and then revert it.
+        root_url = path_to_url(Path().as_posix())
+        replaced = "{root:uri}" in line
+        if replaced:
+            line = line.replace("{root:uri}", root_url)
         try:
-            package_req = PackageRequirement(line)  # type: ignore
+            pkg_req = parse_as_pkg_requirement(line)
         except InvalidRequirement as e:
             m = _file_req_re.match(line)
             if m is None:
-                raise RequirementError(str(e)) from None
-            r = FileRequirement.create(**m.groupdict())
+                raise RequirementError(f"{line}: {e}") from None
+            args = m.groupdict()
+            if not line.startswith(".") and not args["url"] and args["path"] and not os.path.exists(args["path"]):
+                raise RequirementError(f"{line}: {e}") from None
+            r = FileRequirement.create(**args)
         else:
-            r = Requirement.from_pkg_requirement(package_req)
+            r = Requirement.from_pkg_requirement(pkg_req)
+        if replaced:
+            assert isinstance(r, FileRequirement)
+            r.url = r.url.replace(root_url, "{root:uri}")
+            r.path = Path(get_relative_path(r.url) or "")
 
     if editable:
-        if r.is_vcs or r.is_file_or_url and r.is_local_dir:  # type: ignore
+        if r.is_vcs or r.is_file_or_url and r.is_local_dir:  # type: ignore[attr-defined]
             assert isinstance(r, FileRequirement)
             r.editable = True
         else:
-            raise RequirementError(
-                "Editable requirement is only supported for VCS link"
-                " or local directory."
-            )
+            raise RequirementError(f"{line}: Editable requirement is only supported for VCS link or local directory.")
     return r
