@@ -1,10 +1,12 @@
 import os
+import platform
 import re
 import shutil
 import sys
 from unittest.mock import ANY
 
 import pytest
+import shellingham
 
 from pdm.cli.commands.venv import backends
 from pdm.cli.commands.venv.utils import get_venv_prefix
@@ -17,8 +19,11 @@ def with_pip(request):
 
 @pytest.fixture()
 def fake_create(monkeypatch):
-    def fake_create(self, location, *args):
-        location.mkdir(parents=True)
+    def fake_create(self, location, *args, prompt=None):
+        bin_dir = "Scripts" if sys.platform == "win32" else "bin"
+        suffix = ".exe" if sys.platform == "win32" else ""
+        (location / bin_dir).mkdir(parents=True)
+        (location / bin_dir / f"python{suffix}").touch()
 
     monkeypatch.setattr(backends.VirtualenvBackend, "perform_create", fake_create)
     monkeypatch.setattr(backends.VenvBackend, "perform_create", fake_create)
@@ -26,86 +31,102 @@ def fake_create(monkeypatch):
 
 
 @pytest.mark.usefixtures("fake_create")
-def test_venv_create(invoke, project):
-    project.project_config.pop("python.path", None)
+def test_venv_create(pdm, project):
+    project._saved_python = None
     project.project_config["venv.in_project"] = False
-    result = invoke(["venv", "create"], obj=project)
+    result = pdm(["venv", "create"], obj=project)
     assert result.exit_code == 0, result.stderr
-    venv_path = re.match(
-        r"Virtualenv (.+) is created successfully", result.output
-    ).group(1)
+    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
     assert os.path.exists(venv_path)
-    assert "python.path" not in project.project_config
+    assert not project._saved_python
 
 
 @pytest.mark.usefixtures("fake_create")
-def test_venv_create_in_project(invoke, project):
+def test_venv_create_in_project(pdm, project):
     project.project_config["venv.in_project"] = True
-    invoke(["venv", "create"], obj=project, strict=True)
-    invoke(["venv", "create"], obj=project, strict=True)
+    pdm(["venv", "create"], obj=project, strict=True)
     venv_path = project.root / ".venv"
     assert venv_path.exists()
-    assert len(os.listdir(project.root / "venvs")) == 1
+    result = pdm(["venv", "create"], obj=project)
+    assert result.exit_code == 1
+    assert "is not empty" in result.stderr
 
 
 @pytest.mark.usefixtures("fake_create")
-def test_venv_list(invoke, project):
-    project.project_config["venv.in_project"] = False
-    result = invoke(["venv", "create"], obj=project)
-    assert result.exit_code == 0, result.stderr
-    venv_path = re.match(
-        r"Virtualenv (.+) is created successfully", result.output
-    ).group(1)
+def test_venv_create_other_location(pdm, project):
+    pdm(["venv", "-p", project.root.as_posix(), "create"], strict=True)
+    venv_path = project.root / ".venv"
+    assert venv_path.exists()
+    result = pdm(["venv", "-p", project.root.as_posix(), "create"])
+    assert result.exit_code == 1
+    assert "is not empty" in result.stderr
 
-    result = invoke(["venv", "list"], obj=project)
+
+@pytest.mark.usefixtures("fake_create")
+def test_venv_show_path(pdm, project):
+    project.project_config["venv.in_project"] = True
+    pdm(["venv", "create"], obj=project, strict=True)
+    pdm(["venv", "create", "--name", "test"], obj=project, strict=True)
+    result = pdm(["venv", "--path", "in-project"], obj=project, strict=True)
+    assert result.output.strip() == str(project.root / ".venv")
+    result = pdm(["venv", "--path", "test"], obj=project)
+    assert result.exit_code == 0
+    result = pdm(["venv", "--path", "foo"], obj=project)
+    assert result.exit_code == 1
+
+
+@pytest.mark.usefixtures("fake_create")
+def test_venv_list(pdm, project):
+    project.project_config["venv.in_project"] = False
+    result = pdm(["venv", "create"], obj=project)
+    assert result.exit_code == 0, result.stderr
+    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
+
+    result = pdm(["venv", "list"], obj=project)
     assert result.exit_code == 0, result.stderr
     assert venv_path in result.output
 
 
 @pytest.mark.usefixtures("fake_create")
-def test_venv_remove(invoke, project):
+def test_venv_remove(pdm, project):
     project.project_config["venv.in_project"] = False
-    result = invoke(["venv", "create"], obj=project)
+    result = pdm(["venv", "create"], obj=project)
     assert result.exit_code == 0, result.stderr
-    venv_path = re.match(
-        r"Virtualenv (.+) is created successfully", result.output
-    ).group(1)
+    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
     key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]
 
-    result = invoke(["venv", "remove", "non-exist"], obj=project)
+    result = pdm(["venv", "remove", "non-exist"], obj=project)
     assert result.exit_code != 0
 
-    result = invoke(["venv", "remove", "-y", key], obj=project)
+    result = pdm(["venv", "remove", "-y", key], obj=project)
     assert result.exit_code == 0, result.stderr
 
     assert not os.path.exists(venv_path)
 
 
 @pytest.mark.usefixtures("fake_create")
-def test_venv_recreate(invoke, project):
+def test_venv_recreate(pdm, project):
     project.project_config["venv.in_project"] = False
-    result = invoke(["venv", "create"], obj=project)
+    result = pdm(["venv", "create"], obj=project)
     assert result.exit_code == 0, result.stderr
 
-    result = invoke(["venv", "create"], obj=project)
+    result = pdm(["venv", "create"], obj=project)
     assert result.exit_code != 0
 
-    result = invoke(["venv", "create", "-f"], obj=project)
+    result = pdm(["venv", "create", "-f"], obj=project)
     assert result.exit_code == 0, result.stderr
 
 
 @pytest.mark.usefixtures("venv_backends")
-def test_venv_activate(invoke, mocker, project):
+def test_venv_activate(pdm, mocker, project):
     project.project_config["venv.in_project"] = False
-    result = invoke(["venv", "create"], obj=project)
+    result = pdm(["venv", "create"], obj=project)
     assert result.exit_code == 0, result.stderr
-    venv_path = re.match(
-        r"Virtualenv (.+) is created successfully", result.output
-    ).group(1)
+    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
     key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]
 
     mocker.patch("shellingham.detect_shell", return_value=("bash", None))
-    result = invoke(["venv", "activate", key], obj=project)
+    result = pdm(["venv", "activate", key], obj=project)
     assert result.exit_code == 0, result.stderr
     backend = project.config["venv.backend"]
 
@@ -113,54 +134,81 @@ def test_venv_activate(invoke, mocker, project):
         assert result.output.startswith("conda activate")
     else:
         assert result.output.strip("'\"\n").endswith("activate")
-        assert result.output.startswith("source")
+        if platform.system() == "Windows":
+            assert not result.output.startswith("source")
+            assert not result.output.startswith("'")
+        else:
+            assert result.output.startswith("source")
 
 
 @pytest.mark.usefixtures("venv_backends")
-def test_venv_activate_custom_prompt(invoke, mocker, project):
+def test_venv_activate_custom_prompt(pdm, mocker, project):
     project.project_config["venv.in_project"] = False
     creator = mocker.patch("pdm.cli.commands.venv.backends.Backend.create")
-    result = invoke(["venv", "create"], obj=project)
+    result = pdm(["venv", "create"], obj=project)
     assert result.exit_code == 0, result.stderr
     creator.assert_called_once_with(
-        None, [], False, False, project.project_config["venv.prompt"], False
+        None, [], False, False, prompt=project.project_config["venv.prompt"], with_pip=False
     )
 
 
-def test_venv_activate_project_without_python(invoke, project):
-    project.project_config.pop("python.path", None)
-    result = invoke(["venv", "activate"], obj=project)
+def test_venv_activate_project_without_python(pdm, project):
+    project._saved_python = None
+    result = pdm(["venv", "activate"], obj=project)
     assert result.exit_code != 0
     assert "The project doesn't have a saved python.path" in result.stderr
 
 
 @pytest.mark.usefixtures("fake_create")
-def test_venv_activate_error(invoke, project):
+def test_venv_activate_error(pdm, project):
     project.project_config["venv.in_project"] = False
-    result = invoke(["venv", "create"], obj=project, strict=True)
+    result = pdm(["venv", "create"], obj=project, strict=True)
 
-    result = invoke(["venv", "activate", "foo"], obj=project)
+    result = pdm(["venv", "activate", "foo"], obj=project)
     assert result.exit_code != 0
     assert "No virtualenv with key" in result.stderr
 
-    project.project_config["python.path"] = os.path.abspath("fake/bin/python")
-    result = invoke(["venv", "activate"], obj=project)
-    print(project.project_config.get("python.path"))
+    project._saved_python = os.path.abspath("fake/bin/python")
+    result = pdm(["venv", "activate"], obj=project)
     assert result.exit_code != 0, result.output + result.stderr
     assert "Can't activate a non-venv Python" in result.stderr
 
 
+@pytest.mark.usefixtures("venv_backends")
+def test_venv_activate_no_shell(pdm, mocker, project):
+    project.project_config["venv.in_project"] = False
+    result = pdm(["venv", "create"], obj=project)
+    assert result.exit_code == 0, result.stderr
+    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
+    key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]
+
+    mocker.patch("shellingham.detect_shell", side_effect=shellingham.ShellDetectionFailure())
+    result = pdm(["venv", "activate", key], obj=project)
+    assert result.exit_code == 0, result.stderr
+    backend = project.config["venv.backend"]
+
+    if backend == "conda":
+        assert result.output.startswith("conda activate")
+    else:
+        assert result.output.strip("'\"\n").endswith("activate")
+        if platform.system() == "Windows":
+            assert not result.output.startswith("source")
+            assert not result.output.startswith("'")
+        else:
+            assert result.output.startswith("source")
+
+
 @pytest.mark.usefixtures("fake_create")
 @pytest.mark.parametrize("keep_pypackages", [True, False])
-def test_venv_auto_create(invoke, mocker, project, keep_pypackages):
+def test_venv_auto_create(pdm, mocker, project, keep_pypackages):
     creator = mocker.patch("pdm.cli.commands.venv.backends.Backend.create")
-    del project.project_config["python.path"]
+    project._saved_python = None
     if keep_pypackages:
         project.root.joinpath("__pypackages__").mkdir(exist_ok=True)
     else:
         shutil.rmtree(project.root / "__pypackages__", ignore_errors=True)
     project.project_config["python.use_venv"] = True
-    invoke(["install"], obj=project)
+    pdm(["install", "--no-self"], obj=project)
     if keep_pypackages:
         creator.assert_not_called()
     else:
@@ -168,30 +216,26 @@ def test_venv_auto_create(invoke, mocker, project, keep_pypackages):
 
 
 @pytest.mark.usefixtures("fake_create")
-def test_venv_purge(invoke, project):
+def test_venv_purge(pdm, project):
     project.project_config["venv.in_project"] = False
-    result = invoke(["venv", "purge"], obj=project)
+    result = pdm(["venv", "purge"], obj=project)
     assert result.exit_code == 0, result.stderr
 
-    result = invoke(["venv", "create"], obj=project)
+    result = pdm(["venv", "create"], obj=project)
     assert result.exit_code == 0, result.stderr
-    venv_path = re.match(
-        r"Virtualenv (.+) is created successfully", result.output
-    ).group(1)
-    result = invoke(["venv", "purge"], input="y", obj=project)
+    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
+    result = pdm(["venv", "purge"], input="y", obj=project)
     assert result.exit_code == 0, result.stderr
     assert not os.path.exists(venv_path)
 
 
 @pytest.mark.usefixtures("fake_create")
-def test_venv_purge_force(invoke, project):
+def test_venv_purge_force(pdm, project):
     project.project_config["venv.in_project"] = False
-    result = invoke(["venv", "create"], obj=project)
+    result = pdm(["venv", "create"], obj=project)
     assert result.exit_code == 0, result.stderr
-    venv_path = re.match(
-        r"Virtualenv (.+) is created successfully", result.output
-    ).group(1)
-    result = invoke(["venv", "purge", "-f"], obj=project)
+    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
+    result = pdm(["venv", "purge", "-f"], obj=project)
     assert result.exit_code == 0, result.stderr
     assert not os.path.exists(venv_path)
 
@@ -201,14 +245,12 @@ user_options = [("none", True), ("0", False), ("all", False)]
 
 @pytest.mark.usefixtures("venv_backends")
 @pytest.mark.parametrize("user_choices, is_path_exists", user_options)
-def test_venv_purge_interactive(invoke, user_choices, is_path_exists, project):
+def test_venv_purge_interactive(pdm, user_choices, is_path_exists, project):
     project.project_config["venv.in_project"] = False
-    result = invoke(["venv", "create"], obj=project)
+    result = pdm(["venv", "create"], obj=project)
     assert result.exit_code == 0, result.stderr
-    venv_path = re.match(
-        r"Virtualenv (.+) is created successfully", result.output
-    ).group(1)
-    result = invoke(["venv", "purge", "-i"], input=user_choices, obj=project)
+    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
+    result = pdm(["venv", "purge", "-i"], input=user_choices, obj=project)
     assert result.exit_code == 0, result.stderr
     assert os.path.exists(venv_path) == is_path_exists
 
@@ -252,6 +294,7 @@ def test_venv_backend_create(project, mocker, with_pip):
 
 
 def test_conda_backend_create(project, mocker, with_pip):
+    assert project.python
     backend = backends.CondaBackend(project, "3.8")
     assert backend.ident == "3.8"
     mock_call = mocker.patch("subprocess.check_call")
